# Copyright 2018 Zhao Xingyu & An Yuexuan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# -*- coding: utf-8 -*-


# CliffWalking环境
class CliffWalking:
    def __init__(self):
        self.state = 0
        self.n_actions = 4
        self.n_states = 48

    def reset(self):
        self.state = 0
        return self.state

    def step(self, a):
        if a == 0:  # 往左走
            if not (self.state % 12 == 0):
                self.state -= 1
        elif a == 1:  # 往上走
            if not (self.state > 35):
                self.state += 12
        elif a == 2:  # 向右走
            if not ((self.state + 1) % 12 == 0):
                self.state += 1
        else:  # 往下走
            if not (self.state < 12):
                self.state -= 12
        if 0 < self.state < 11:
            reward = -100
        else:
            reward = -1
        if 0 < self.state < 12:
            done = True
        else:
            done = False
        return self.state, reward, done, None


# FrozenLake环境
class FrozenLake:
    def __init__(self, n):
        self.state = 0
        self.n = n
        self.n_actions = 4
        self.n_states = n * n

    def reset(self):
        self.state = 0
        return self.state

    def step(self, a):
        # 执行动作
        if a == 0:  # 往左走
            if not (self.state % self.n == 0):
                self.state -= 1
        elif a == 1:  # 往上走
            if not (self.state < self.n):
                self.state -= self.n
        elif a == 2:  # 向右走
            if not ((self.state + 1) % self.n == 0):
                self.state += 1
        else:  # 向下走
            if not (self.state + 1 > self.n * (self.n - 1)):
                self.state += self.n
        # 其他操作
        if self.n == 4:
            if self.state == 5 or self.state == 7 or self.state == 11 or self.state == 12:
                reward = -10
                done = True
            elif self.state == 15:
                reward = 10
                done = True
            else:
                reward = -1
                done = False
            return self.state, reward, done, None
        elif self.n == 5:
            if self.state == 6 or self.state == 8 or self.state == 10 or self.state == 19 or self.state == 22:
                reward = -15
                done = True
            elif self.state == 24:
                reward = 15
                done = True
            else:
                reward = -1
                done = False
            return self.state, reward, done, None
        elif self.n == 6:
            if self.state == 13 or self.state == 14 or self.state == 16 or self.state == 25 or self.state == 27 or self.state == 29 or self.state == 30:
                reward = -20
                done = True
            elif self.state == 35:
                reward = 20
                done = True
            else:
                reward = -1
                done = False
            return self.state, reward, done, None
        elif self.n == 7:
            if self.state == 16 or self.state == 20 or self.state == 21 or self.state == 25 or self.state == 29 or self.state == 32 or self.state == 39 or self.state == 41 or self.state == 44:
                reward = -25
                done = True
            elif self.state == 48:
                reward = 25
                done = True
            else:
                reward = -1
                done = False
            return self.state, reward, done, None
        elif self.n == 8:
            if self.state == 16 or self.state == 25 or self.state == 27 or self.state == 31 or self.state == 37 or self.state == 38 or self.state == 42 or self.state == 44 or self.state == 54 or self.state == 61:
                reward = -30
                done = True
            elif self.state == 63:
                reward = 30
                done = True
            else:
                reward = -1
                done = False
            return self.state, reward, done, None
        elif self.n == 9:
            if self.state == 28 or self.state == 38 or self.state == 40 or self.state == 44 or self.state == 51 or self.state == 52 or self.state == 54 or self.state == 55 or self.state == 57 or self.state == 59 or self.state == 70 or self.state == 78:
                reward = -35
                done = True
            elif self.state == 80:
                reward = 35
                done = True
            else:
                reward = -1
                done = False
            return self.state, reward, done, None


# GridWorld环境
class GridWorld:
    def __init__(self, n):
        self.state = 0
        self.n = n
        self.n_actions = 4
        self.n_states = n * n

    def reset(self):
        self.state = 0
        return self.state

    def step(self, a):
        # 执行动作
        if a == 0:  # 往左走
            if self.state == (self.n - 2) * (self.n + 1) / 2:
                reward = -5
            elif not (self.state % self.n == 0):
                reward = -1
                self.state -= 1
            else:
                reward = -5
        elif a == 1:  # 往上走
            if self.n ** 2 / 2 - 1 < self.state < (self.n + 1) * self.n / 2 - 1 or (
                    self.n + 1) * self.n / 2 < self.state < (self.n + 2) * self.n / 2:
                reward = -5
            elif not (self.state < self.n):
                reward = -1
                self.state -= self.n
            else:
                reward = -5
        elif a == 2:  # 向右走
            if self.state == (self.n - 2) * (self.n + 1) / 2 + 1:
                reward = -5
            elif not ((self.state + 1) % self.n == 0):
                reward = -1
                self.state += 1
            else:
                reward = -5
        else:  # 向下走
            if self.n * (self.n - 4) / 2 - 1 < self.state < self.n * (self.n - 3) / 2 - 1 or (
                    self.n - 3) * self.n / 2 < self.state < (self.n - 2) * self.n / 2:
                reward = -5
            elif not (self.state + 1 > self.n * (self.n - 1)):
                reward = -1
                self.state += self.n
            else:
                reward = -5
        # 其他操作
        done = False
        if self.state == self.n ** 2 - 1:
            reward = self.n ** 2
            done = True
        return self.state, reward, done, None


# WindyGridWorld环境
class WindyGridWorld:
    def __init__(self):
        self.state = 30
        self.n_actions = 4
        self.n_states = 70

    def reset(self):
        self.state = 30
        return self.state

    def step(self, a):
        # 风力处理
        if self.state % 10 == 3 or self.state % 10 == 4 or self.state % 10 == 5 or self.state % 10 == 8:
            self.state += 10
        elif self.state % 10 == 6 or self.state % 10 == 7:
            self.state += 20
        if self.state > 70:
            self.state = 60 + self.state % 10
        # 执行动作
        if a == 0:  # 往左走
            if not self.state % 10 == 0:
                self.state -= 1
        elif a == 1:  # 往上走
            if not self.state >= 60:
                self.state += 10
        elif a == 2:  # 向右走
            if not self.state % 10 == 9:
                self.state += 1
        else:  # 向下走
            if not self.state < 10:
                self.state -= 10
        # 其他操作
        if self.state == 37:
            reward = 0
            done = True
        else:
            reward = -1
            done = False
        return self.state, reward, done, None

    def render(self):
        for i in range(6, -1, -1):
            a = []
            for j in range(10):
                a.append('x')
            if i == 3:
                a[0] = 'S'
                a[7] = 'G'
            if self.state // 10 == i:
                a[self.state % 10] = 'o'
            print(' '.join(a))



# Copyright 2018 Zhao Xingyu & An Yuexuan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
import numpy as np
from sklearn.model_selection import StratifiedKFold, train_test_split
import warnings
from sklearn.metrics import precision_recall_fscore_support, confusion_matrix


# 准确率
def accuracy_score(prediction, label):
    '''
    :param prediction: 预测类别或one-hot编码
    :param label: 实际类别或one-hot编码
    :return: 准确率
    '''
    prediction, label = np.array(prediction), np.array(label)
    assert len(prediction.shape) == 1 or len(prediction.shape) == 2  # 输出值形状错误
    assert len(label.shape) == 1 or len(label.shape) == 2  # 真实值形状错误
    if len(prediction.shape) == 2:
        if prediction.shape[1] == 1:
            prediction = prediction.squeeze()
        else:
            prediction = np.argmax(prediction, 1)
    if len(label.shape) == 2:
        if label.shape[1] == 1:
            label = label.squeeze()
        else:
            label = np.argmax(label, 1)
    return np.mean(np.equal(prediction, label))


# 均方误差
def MSE(prediction, y, sample_importance=None):
    '''
    :param prediction: 预测值
    :param y: 真实值
    :param sample_importance: 样本重要性权重
    :return: 均方误差
    '''
    prediction, y = np.array(prediction).squeeze(), np.array(y).squeeze()
    assert len(prediction.shape) == 1 and len(y.shape) == 1  # 预测值或真实值形状错误
    if sample_importance is None:
        return np.mean(np.square(prediction - y))
    else:
        sample_importance = np.array(sample_importance)
        assert sample_importance.shape[0] == prediction.shape[0]  # 重要性权值形状错误
        return np.mean(sample_importance * np.square(prediction - y))


# 均方根误差
def RMSE(prediction, y, sample_importance=None):
    '''
    :param prediction: 预测值
    :param y: 真实值
    :param sample_importance: 样本重要性权重
    :return: 均方根误差
    '''
    prediction, y = np.array(prediction).squeeze(), np.array(y).squeeze()
    assert len(prediction.shape) == 1 and len(y.shape) == 1  # 预测值或真实值形状错误
    if sample_importance is None:
        return np.sqrt(np.mean(np.square(prediction - y)))
    else:
        sample_importance = np.array(sample_importance)
        assert sample_importance.shape[0] == prediction.shape[0]  # 重要性权值形状错误
        return np.sqrt(np.mean(sample_importance * np.square(prediction - y)))


# 平均绝对误差
def MAE(prediction, y, sample_importance=None):
    '''
    :param prediction: 预测值
    :param y: 真实值
    :param sample_importance: 样本重要性权重
    :return:
    '''
    prediction, y = np.array(prediction).squeeze(), np.array(y).squeeze()
    assert len(prediction.shape) == 1 and len(y.shape) == 1  # 预测值或真实值形状错误
    if sample_importance is None:
        return np.mean(np.abs(prediction - y))
    else:
        sample_importance = np.array(sample_importance)
        assert sample_importance.shape[0] == prediction.shape[0]  # 重要性权值形状错误
        return np.mean(sample_importance * np.abs(prediction - y))


# 误差平方和
def SSE(prediction, y, sample_importance=None):
    '''
    :param prediction: 预测值
    :param y: 真实值
    :param sample_importance: 样本重要性权重
    :return:
    '''
    prediction, y = np.array(prediction).squeeze(), np.array(y).squeeze()
    assert len(prediction.shape) == 1 and len(y.shape) == 1  # 预测值或真实值形状错误
    if sample_importance is None:
        return np.sum(np.square(prediction - y))
    else:
        sample_importance = np.array(sample_importance)
        assert sample_importance.shape[0] == prediction.shape[0]  # 重要性权值形状错误
        return np.sum(sample_importance * np.square(prediction - y))


# 总平方和
def SST(y, sample_importance=None):
    '''
    :param y: 真实值
    :param sample_importance: 样本重要性权重
    :return: 总平方和
    '''
    y = np.array(y)
    assert len(y.shape) == 1  # 真实值形状错误
    if sample_importance is None:
        return np.sum(np.square(y - np.mean(y)))
    else:
        sample_importance = np.array(sample_importance)
        assert sample_importance.shape[0] == y.shape[0]  # 重要性权值形状错误
        return np.sum(sample_importance * np.square(y - np.mean(y)))


# 回归平方和
def SSR(prediction, y, sample_importance=None):
    '''
    :param prediction: 预测值
    :param y: 真实值
    :param sample_importance: 样本重要性权重
    :return: 回归平方和
    '''
    prediction, y = np.array(prediction).squeeze(), np.array(y).squeeze()
    assert len(prediction.shape) == 1 and len(y.shape) == 1  # 预测值或真实值形状错误
    if sample_importance is None:
        return np.sum(np.square(prediction - np.mean(y)))  # Total sum of squares
    else:
        sample_importance = np.array(sample_importance)
        assert sample_importance.shape[0] == prediction.shape[0]  # 重要性权值形状错误
        return np.sum(sample_importance * np.square(prediction - np.mean(y)))


# 确定系数
def R_square(prediction, y, sample_importance=None):
    '''
    :param prediction: 预测值
    :param y: 真实值
    :param sample_importance: 样本重要性权重
    :return: 确定系数
    '''
    return 1 - SSE(prediction, y, sample_importance) / SST(y, sample_importance)


# 皮尔森相关系数
def PC(prediction, y):
    '''
    :param prediction: 预测值
    :param y: 真实值
    :return:
    '''
    prediction, y = np.array(prediction).squeeze(), np.array(y).squeeze()
    assert len(prediction.shape) == 1 and len(y.shape) == 1  # 预测值或真实值形状错误
    n = y.shape[0]
    return (n * np.sum(prediction * y) - np.sum(prediction) * np.sum(y)) / (
            np.sqrt(n * np.sum(prediction ** 2) - np.sum(prediction) ** 2) * np.sqrt(
        n * np.sum(y ** 2) - np.sum(y) ** 2))


# K折交叉验证
def cross_val_score(estimator, x, y, k=10, verbose=True, random_state=None, **kwargs):
    '''
    :param estimator: 待评价的模型
    :param x: 样本数据
    :param y: 样本标签
    :param k: K折交叉验证中的K值
    :param verbose: 是否显示验证过程
    :param random_state: 数据集分割的随机数种子
    :param kwargs: estimator.fit()的参数
    :return: k次验证的准确率一维数组
    '''
    x, y = np.array(x), np.array(y)
    if random_state is None:
        folder = StratifiedKFold(k, True)
    else:
        folder = StratifiedKFold(k, True, random_state)
    scores = []
    for i, (train_index, test_index) in enumerate(folder.split(x, y)):
        estimator.fit(x[train_index], y[train_index], **kwargs)
        score = estimator.score(x[test_index], y[test_index])
        scores.append(score)
        if verbose:
            print('第%d次交叉验证完成，得分为%.4f' % (i + 1, score))
    scores = np.array(scores)
    return scores


# 留p法交叉验证
def leave_p_score(estimator, x, y, p=1, verbose=True, **kwargs):
    '''
    :param estimator: 待评价的模型
    :param x: 样本数据
    :param y: 样本标签
    :param p: 留p法交叉验证中的p值
    :param verbose: 是否显示验证过程
    :param kwargs: estimator.fit()的参数
    :return: len(x)//p次验证的准确率一维数组
    '''
    x, y = np.array(x), np.array(y)
    if x.shape[0] < p:
        warnings.warn('交叉验证参数错误，不执行操作！')
        return None
    epoch = x.shape[0] // p
    index = np.arange(x.shape[0])
    np.random.shuffle(index)
    scores = []
    for i in range(epoch):
        test_index = slice(i * p, (i + 1) * p)
        train_index = np.delete(index, test_index)
        estimator.fit(x[train_index], y[train_index], **kwargs)
        score = estimator.score(x[test_index], y[test_index])
        scores.append(score)
        if verbose:
            print('第%d次交叉验证完成，得分为%.4f' % (i + 1, score))
    scores = np.array(scores)
    return scores


# Hold-Out检验
def hold_out_score(estimator, x, y, test_size=0.25, shuffle=True, stratify=None, random_state=0, **kwargs):
    '''
    :param estimator: 待评价的模型
    :param x: 样本数据
    :param y: 样本标签
    :param test_size: 测试数据比率
    :param shuffle: 是否随机化
    :param stratify: 是否分层采样(一般输入y)
    :param random_state: 数据集分割的随机数种子
    :param kwargs: estimator.fit()的参数
    :return: 准确率
    '''
    x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=test_size, shuffle=shuffle, stratify=stratify,
                                                        random_state=random_state)
    estimator.fit(x_train, y_train, **kwargs)
    return estimator.score(x_test, y_test)


# 分类报告
def class_report(y_true, y_pred, method='weighted'):
    '''
    :param y_true: 真实标签
    :param y_pred: 预测标签
    :param method: 报告方式，包括'weighted'、'micro'、'macro'
    :return: 分类报告字典
    '''
    y_true, y_pred = np.array(y_true), np.array(y_pred)
    accuracy = np.mean(y_true == y_pred)
    cm = confusion_matrix(y_true, y_pred)  # 混淆矩阵
    _, _, _, num = precision_recall_fscore_support(y_true, y_pred)  # num为每一类的样本数目列表
    n_classes = len(num)
    n_samples = len(y_true)
    TP, FN, FP, TN = np.zeros([n_classes], int), np.zeros([n_classes], int), np.zeros([n_classes], int), np.zeros(
        [n_classes], int)  # 每一类的TP、FN、FP、TN
    precision, recall, recall_, F1_score, G_mean = np.zeros([n_classes]), np.zeros([n_classes]), np.zeros(
        [n_classes]), np.zeros([n_classes]), np.zeros([n_classes])  # 每一类的精确度、召回率、负类召回率、F1值、G-mean值
    # 微平均
    if method == 'micro':
        for i in range(n_classes):
            TP[i] = cm[i][i]
            FP[i] = np.sum(cm[:, i]) - TP[i]
            FN[i] = np.sum(cm[i]) - TP[i]
            TN[i] = n_samples - TP[i] - FP[i] - FN[i]
        TP, FP, FN, TN = np.mean(TP), np.mean(FP), np.mean(FN), np.mean(TN)
        if TP == 0 and FP == 0:
            precision = 0
        else:
            precision = TP / (TP + FP)
        if TP == 0 and FN == 0:
            recall = 0
        else:
            recall = TP / (TP + FN)
        if TN == 0 and FP == 0:
            recall_ = 0
        else:
            recall_ = TN / (TN + FP)
        if recall == 0 and precision == 0:
            F1_score = 0
        else:
            F1_score = 2 * precision * recall / (precision + recall)
        if recall == 0 and recall_ == 0:
            G_mean = 0
        else:
            G_mean = np.sqrt(recall * recall_)
        return {'accuracy': accuracy, 'precision': precision, 'recall': recall, 'F1_score': F1_score, 'G_mean': G_mean}

    # 宏平均或加权平均
    elif method == 'macro' or method == 'weighted':
        for i in range(n_classes):
            TP[i] = cm[i][i]
            FP[i] = np.sum(cm[:, i]) - TP[i]
            FN[i] = np.sum(cm[i]) - TP[i]
            TN[i] = n_samples - TP[i] - FP[i] - FN[i]
            if TP[i] == 0 and FP[i] == 0:
                precision[i] = 0
            else:
                precision[i] = TP[i] / (TP[i] + FP[i])
            if TP[i] == 0 and FN[i] == 0:
                recall[i] = 0
            else:
                recall[i] = TP[i] / (TP[i] + FN[i])
            if TN[i] == 0 and FP[i] == 0:
                recall_[i] = 0
            else:
                recall_[i] = TN[i] / (TN[i] + FP[i])
            if recall[i] == 0 and precision[i] == 0:
                F1_score[i] = 0
            else:
                F1_score[i] = 2 * precision[i] * recall[i] / (precision[i] + recall[i])
            if recall[i] == 0 and recall_[i] == 0:
                G_mean[i] = 0
            else:
                G_mean[i] = np.sqrt(recall[i] * recall_[i])
        if method == 'macro':
            precision = np.mean(precision)
            recall = np.mean(recall)
            F1_score = np.mean(F1_score)
            G_mean = np.mean(G_mean)
            return {'accuracy': accuracy, 'precision': precision, 'recall': recall, 'F1_score': F1_score,
                    'G_mean': G_mean}
        else:
            precision = np.sum(precision * num) / n_samples
            recall = np.sum(recall * num) / n_samples
            F1_score = np.sum(F1_score * num) / n_samples
            G_mean = np.sum(G_mean * num) / n_samples
            return {'accuracy': accuracy, 'precision': precision, 'recall': recall, 'F1_score': F1_score,
                    'G_mean': G_mean}

    else:
        print('Classification report type error!')



# -*- coding: utf-8 -*-
# Copyright 2018 Zhao Xingyu & An Yuexuan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings
from .utils import to_categorical
from .nn.activations import *
from sklearn.metrics import r2_score


# 核方法（A:[m1,n],B:[m2,n],K:[m1,m2]）
def Kernel(A, B=None, kernel='linear', gamma=0.1, q=0.1, degree=3):
    '''
    :param A: 矩阵1 ，形状为[m1,n]
    :param B: 矩阵2，形状为[m2,n]，取值为None时B=A.copy()
    :param kernel: 核的类型，['linear','polynomial'('poly'),'sigmoid','rbf','laplace']
    :param gamma: 多项式核、Sigmoid核、高斯核、拉普拉斯核中的参数
    :param q: 多项式核、Sigmoid核、高斯核中的参数
    :param degree: 多项式核中的参数
    :return: 核运算得到的矩阵[m1,m2]
    '''
    A = np.array(A, np.float)
    if B is None:
        B = A.copy()
    B = np.array(B, np.float)
    if kernel == 'linear':  # 线性核
        my_kernel = A.dot(B.T)
    elif kernel == 'polynomial' or kernel == 'poly':  # 多项式核
        my_kernel = (gamma * A.dot(B.T) + q) ** degree
    elif kernel == 'sigmoid':  # Sigmoid核
        my_kernel = np.tanh(gamma * A.dot(B.T) - q)
    elif kernel == 'rbf':  # 高斯核
        rA = np.sum(np.square(A), 1, keepdims=True)
        rB = np.sum(np.square(B), 1, keepdims=True)
        sq_dists = rA - 2 * A.dot(B.T) + np.transpose(rB)  # x^2-2*x*y+y^2
        my_kernel = np.exp(-gamma * np.abs(sq_dists))
    elif kernel == 'laplace':  # 拉普拉斯核
        A, B = np.array(A), np.array(B)
        rA, rB = np.expand_dims(A, 1), np.expand_dims(B, 0)
        my_kernel = np.exp(-gamma * np.sum(np.abs(rA - rB), axis=2))
    else:
        print('kernel error!')
        return
    return my_kernel


# 主成分分析
def PCA(x, feature_num=None, svd=False):
    '''
    :param x: 样本矩阵，形状为[样本个数,特征维数]
    :param feature_num: 保留的特征数，为None时等于特征维数
    :param svd: 是否使用svd进行PCA
    :return: 计算后得到的矩阵，形状为[样本个数,feature_num]
    '''
    x = np.array(x, np.float)
    if len(x.shape) != 1 and len(x.shape) != 2:
        warnings.warn('数据维度不正确，不执行操作！')
        return None
    if len(x.shape) == 1:
        x = np.expand_dims(x, 0)
    if feature_num is None:
        feature_num = x.shape[1]
    x -= x.mean(0, keepdims=True)
    if svd:
        U, S, VT = np.linalg.svd(x)
        index_sort = np.argsort(S)  # 对奇异值进行排序
        index = index_sort[-1:-(feature_num + 1):-1]
        return x.dot(VT[index])  # 乘上最大的feature_num个奇异值组成的特征向量
    else:
        eigval, eigvec = np.linalg.eig(x.transpose().dot(x) / x.shape[0])
        index_sort = np.argsort(eigval)  # 对特征值进行排序
        index = index_sort[-1:-(feature_num + 1):-1]
        return x.dot(eigvec[:, index])  # 乘上最大的feature_num个特征组成的特征向量


# 核化主成分分析
def KernelPCA(x, feature_num=None, kernel='rbf', gamma=0.1, q=0.1, degree=3):
    # x：样本
    # feature_num：保留的特征数
    x = np.array(x, np.float)
    if len(x.shape) != 1 and len(x.shape) != 2:
        warnings.warn('数据维度不正确，不执行操作！')
        return None
    if len(x.shape) == 1:
        x = np.expand_dims(x, 0)
    if feature_num is None:
        feature_num = x.shape[1]
    K = Kernel(x, kernel=kernel, gamma=gamma, q=q, degree=degree)
    one_m = np.ones([x.shape[0], x.shape[0]]) / x.shape[0]
    K = K - one_m.dot(K) - K.dot(one_m) + one_m.dot(K).dot(one_m)
    eigval, eigvec = np.linalg.eig(K)
    index_sort = np.argsort(eigval)  # 对特征值进行排序
    index = index_sort[-1:-(feature_num + 1):-1]
    lambdas = eigval[index]
    alphas = eigvec[:, index]  # 乘上最大的feature_num个特征组成的特征向量
    return K.dot(alphas / np.sqrt(lambdas))


# ELM分类器
class ELMClassifier:
    def __init__(self, n_hidden=150, activation='leaky_relu', kernel=None, gamma=0.1, q=0.1, degree=3):
        self.n_hidden = n_hidden  # 不使用核时，隐层节点个数
        self.activation = activation.lower()  # 不使用核时，激活函数
        self.w1, self.b, self.w2 = None, None, None  # 参数值
        self.kernel = kernel  # 核的类型，[None,'linear','polynomial'('poly'),'sigmoid','rbf','laplace']
        self.gamma = gamma  # 多项式核、Sigmoid核、高斯核、拉普拉斯核中的参数
        self.q = q  # 多项式核、Sigmoid核、高斯核中的参数
        self.degree = degree  # 多项式核中的参数
        self.x_train = None  # 训练数据

    def fit(self, x_train, y_train):
        x_train, y_train = np.array(x_train), np.array(y_train)
        self.x_train = x_train.copy()
        if self.kernel == None:
            self.w1 = np.random.uniform(-1, 1, (x_train.shape[1], self.n_hidden))
            self.b = np.random.uniform(-1, 1, self.n_hidden)
            H = eval('%s(x_train.dot(self.w1) + self.b)' % self.activation)
            self.w2 = np.linalg.pinv(H).dot(to_categorical(y_train))
        else:
            H = Kernel(x_train, self.x_train, kernel=self.kernel, gamma=self.gamma, q=self.q, degree=self.degree)
            self.w2 = np.linalg.pinv(H).dot(to_categorical(y_train))

    def predict(self, x_test):
        x_test = np.array(x_test)
        if self.kernel == None:
            x_test = eval(' %s(x_test.dot(self.w1) + self.b).dot(self.w2)' % self.activation)
        else:
            x_test = Kernel(x_test, self.x_train, kernel=self.kernel, gamma=self.gamma, q=self.q,
                            degree=self.degree).dot(self.w2)
        return np.argmax(x_test, axis=1)

    def score(self, x_test, y_test):
        x_test, y_test = np.array(x_test), np.array(y_test)
        y_pred = self.predict(x_test)
        return np.mean(np.equal(y_test, y_pred))

    # 保存模型
    def save(self, filepath):
        np.savez(filepath, n_hidden=self.n_hidden, activation=self.activation, w1=self.w1, b=self.b, w2=self.w2,
                 kernel=self.kernel, gamma=self.gamma, q=self.q, degree=self.degree, x_train=self.x_train)

    # 读取模型
    def load(self, filepath):
        if '.npz' not in filepath:
            filepath = filepath + '.npz'
        params = np.load(filepath)
        self.n_hidden = params['n_hidden']
        self.activation = params['activation']
        self.w1, self.b, self.w2 = params['w1'], params['b'], params['w2']
        self.kernel = params['kernel']
        self.gamma, self.q, self.degree = params['gamma'], params['q'], params['degree']
        self.x_train = params['x_train']


# ELM回归器
class ELMRegressor:
    def __init__(self, n_hidden=150, activation='leaky_relu', kernel=None, gamma=0.1, q=0.1, degree=3):
        self.n_hidden = n_hidden  # 不使用核时，隐层节点个数
        self.activation = activation.lower()  # 不使用核时，激活函数
        self.w1, self.b, self.w2 = None, None, None  # 参数值
        self.kernel = kernel  # 核的类型，[None,'linear','polynomial'('poly'),'sigmoid','rbf','laplace']
        self.gamma = gamma  # 多项式核、Sigmoid核、高斯核、拉普拉斯核中的参数
        self.q = q  # 多项式核、Sigmoid核、高斯核中的参数
        self.degree = degree  # 多项式核中的参数
        self.x_train = None  # 训练数据

    def fit(self, x_train, y_train):
        x_train, y_train = np.array(x_train), np.array(y_train).reshape(-1, 1)
        self.x_train = x_train.copy()
        if self.kernel == None:
            self.w1 = np.random.uniform(-1, 1, (x_train.shape[1], self.n_hidden))
            self.b = np.random.uniform(-1, 1, self.n_hidden)
            H = eval('%s(x_train.dot(self.w1) + self.b)' % self.activation)
            self.w2 = np.linalg.pinv(H).dot(y_train)
        else:
            H = Kernel(x_train, self.x_train, kernel=self.kernel, gamma=self.gamma, q=self.q, degree=self.degree)
            self.w2 = np.linalg.pinv(H).dot(y_train)

    def predict(self, x_test):
        x_test = np.array(x_test)
        if self.kernel == None:
            return eval(' %s(x_test.dot(self.w1) + self.b).dot(self.w2)' % self.activation)
        else:
            return Kernel(x_test, self.x_train, kernel=self.kernel, gamma=self.gamma, q=self.q,
                          degree=self.degree).dot(self.w2)

    def score(self, x_test, y_test):
        x_test, y_test = np.array(x_test), np.array(y_test)
        y_pred = self.predict(x_test)
        return r2_score(y_test, y_pred)

    # 保存模型
    def save(self, filepath):
        np.savez(filepath, n_hidden=self.n_hidden, activation=self.activation, w1=self.w1, b=self.b, w2=self.w2,
                 kernel=self.kernel, gamma=self.gamma, q=self.q, degree=self.degree, x_train=self.x_train)

    # 读取模型
    def load(self, filepath):
        if '.npz' not in filepath:
            filepath = filepath + '.npz'
        params = np.load(filepath)
        self.n_hidden = params['n_hidden']
        self.activation = params['activation']
        self.w1, self.b, self.w2 = params['w1'], params['b'], params['w2']
        self.kernel = params['kernel']
        self.gamma, self.q, self.degree = params['gamma'], params['q'], params['degree']
        self.x_train = params['x_train']





# Using PSO for optimization
from ailearn.Swarm import PSO


def func(x, y):
    return -(x ** 2 + y ** 2)


p = PSO(func=func, param_len=2, x_min=[-5, -5], x_max=[5, 5])
x, y = p.solve()
print(x, y)



import requests
from bs4 import BeautifulSoup
import random

ualist = [
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:23.0) Gecko/20131011 Firefox/23.0",
    "Mozilla/5.0 (X11; Linux i586; rv:31.0) Gecko/20100101 Firefox/31.0",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/31.0.1623.0 Safari/537.36",
    "Mozilla/5.0 (Windows; U; Windows NT 6.0; nb-NO) AppleWebKit/533.18.1 (KHTML, like Gecko) Version/5.0.2 Safari/533.18.5",
    "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/36.0.1985.67 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2227.1 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/35.0.1916.47 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/537.14 (KHTML, like Gecko) Chrome/24.0.1292.0 Safari/537.14",
    "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64; Trident/5.0; SLCC2; Media Center PC 6.0; InfoPath.3; MS-RTC LM 8; Zune 4.7)",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2227.1 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.2; rv:21.0) Gecko/20130326 Firefox/21.0",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/37.0.2049.0 Safari/537.36",
    "Mozilla/5.0 (X11; Linux x86_64; rv:28.0) Gecko/20100101  Firefox/28.0",
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.99 Safari/537.36'
]


def get_html(url, headers=None):
    if headers is not None:
        html = requests.get(url, headers=headers)
    else:
        html = requests.get(url, headers={'User-Agent': random.choice(ualist)})
    html.encoding = 'utf-8'
    return html


def get_soup(html, features='html.parser'):
    '''

    :param html: class:`Response <Response>` object
    :param features: lxml, html.parser or html5lib
    :return:
    '''
    obj = BeautifulSoup(html.text, features=features)
    return obj


def get_soup_from_url(url, headers=None, features='html.parser'):
    return get_soup(get_html(url, headers=headers), features=features)



# Copyright 2018 Zhao Xingyu & An Yuexuan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
# 若干进化算法的Python实现
import numpy as np


# 进化策略
class ES:
    def __init__(self, func=None, param_len=1, size=50, x_min=-10., x_max=10., alpha=0.5):
        self.func = func  # 计算适应性系数的方法
        self.param_len = param_len  # 参数个数
        self.size = size  # 有多少元素
        self.alpha = alpha  # 淘汰率
        self.history = []  # 每次迭代的最优值
        assert np.floor(self.size * (1 - self.alpha)) > 1
        # 最小数值
        if type(x_min) != list:
            self.x_min = [x_min] * self.param_len
        else:
            assert len(x_min) == self.param_len
            self.x_min = x_min
        # 最大数值
        if type(x_max) != list:
            self.x_max = [x_max] * self.param_len
        else:
            assert len(x_max) == self.param_len
            self.x_max = x_max
        self.x = None  # 数值
        self.v = None  # 变化幅度
        self.best_all_x = None  # 全局最优位置
        self.best_all_score = None  # 全局最优分数
        self.score = None  # 得分
        self._init_fit()

    def _init_fit(self):
        self.x = np.zeros([self.size, self.param_len])  # 形状为[微粒个数,参数个数]
        self.v = np.random.uniform(size=[self.size, self.param_len])
        self.score = np.zeros(self.size)
        for i in range(self.size):
            for j in range(self.param_len):
                self.x[i][j] = np.random.uniform(self.x_min[j], self.x_max[j])
        self.best_all_x = np.zeros(self.param_len)  # 全局最优位置
        self.best_all_score = -np.inf  # 全局最优分数

    def solve(self, epoch=50, verbose=False):
        self.history = []
        for _ in range(epoch):  # 一共迭代_次
            self.score = np.zeros(self.size)
            # 计算适应度
            for i in range(self.size):  # 对于第i个微粒
                self.score[i] = self.func(*self.x[i])
            # 更新局部最优值、局部最优位置
            if np.max(self.score) > self.best_all_score:
                self.best_all_score = np.max(self.score)
                self.best_all_x = self.x[np.argmax(self.score)]
            # 优胜劣汰
            for i in range(int(self.size * self.alpha)):
                idx = np.argmin(self.score)
                self.x = np.delete(self.x, idx, axis=0)
                self.v = np.delete(self.v, idx, axis=0)
                self.score = np.delete(self.score, idx, axis=0)
            # 繁殖
            x = np.zeros([self.size, self.param_len])
            v = np.zeros([self.size, self.param_len])
            for i in range(self.size):
                a, b = np.random.choice(self.x.shape[0], 2, False)
                mean = (self.x[a] + self.x[b]) / 2
                v[i] = np.abs(np.random.normal((self.v[a] + self.v[b]) / 2, 0.1))
                x[i] = np.clip(np.random.normal(mean, v[i]), self.x_min, self.x_max)
            self.x = x.copy()
            self.v = v.copy()
            if verbose:
                print('Number of iterations: %i.' % (_ + 1), 'Best fitness: %.4f' % self.best_all_score)
            self.history.append(self.best_all_score)
        self.history = np.array(self.history)
        return self.best_all_x



# Copyright 2018 Zhao Xingyu & An Yuexuan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
import numpy as np
import warnings


# 闵可夫斯基距离
def minkowski_distance(a, b, p):
    '''
    :param a: 向量1
    :param b: 向量2
    :param p: 参数（阶数）
    :return: 闵可夫斯基距离
    '''
    a = np.array(a).squeeze()
    b = np.array(b).squeeze()
    if len(a.shape) != 1 or len(b.shape) != 1:
        warnings.warn('数据维度不为1，不执行操作！')
        return None
    return np.power(np.sum(np.power(np.abs(a - b), p)), 1 / p)


# l1范数
def l1_distance(a, b):
    '''
    :param a: 向量1
    :param b: 向量2
    :return: l1范数
    '''
    return minkowski_distance(a, b, 1)


# 曼哈顿距离
def manhattan_distance(a, b):
    '''
    :param a: 向量1
    :param b: 向量2
    :return: 曼哈顿距离
    '''
    return minkowski_distance(a, b, 1)


# l2范数
def l2_distance(a, b):
    '''
    :param a: 向量1
    :param b: 向量2
    :return: l2范数
    '''
    return minkowski_distance(a, b, 2)


# 欧拉距离
def euclidean_distance(a, b):
    '''
    :param a: 向量1
    :param b: 向量2
    :return: 欧拉距离
    '''
    return minkowski_distance(a, b, 2)


# 切比雪夫距离
def chebyshev_distance(a, b):
    '''
    :param a: 向量1
    :param b: 向量2
    :return: 切比雪夫距离
    '''
    a = np.array(a).squeeze()
    b = np.array(b).squeeze()
    if len(a.shape) != 1 or len(b.shape) != 1:
        warnings.warn('数据维度不为1，不执行操作！')
        return None
    return np.max(np.abs(a - b))


# 夹角余弦
def cosine(a, b):
    '''
    :param a: 向量1
    :param b: 向量2
    :return: 夹角余弦
    '''
    a = np.array(a).squeeze()
    b = np.array(b).squeeze()
    if len(a.shape) != 1 or len(b.shape) != 1:
        warnings.warn('数据维度不为1，不执行操作！')
        return None
    return a.dot(b) / (np.linalg.norm(a, 2) * np.linalg.norm(b, 2))


# 汉明距离
def hamming_distance(a, b):
    '''
    :param a: 向量1
    :param b: 向量2
    :return: 汉明距离
    '''
    a = np.array(a, np.str).squeeze()
    b = np.array(b, np.str).squeeze()
    if len(a.shape) != 1 or len(b.shape) != 1:
        warnings.warn('数据维度不为1，不执行操作！')
        return None
    return np.sum(a != b)


# 杰拉德相似系数
def jaccard_similarity_coefficient(a, b):
    '''
    :param a: 向量1
    :param b: 向量2
    :return: 杰拉德相似系数
    '''
    a = set(a)
    b = set(b)
    return len(a.intersection(b)) / len(a.union(b))


# 杰拉德距离
def jaccard_distance(a, b):
    '''
    :param a: 向量1
    :param b: 向量2
    :return: 杰拉德距离
    '''
    return 1 - jaccard_similarity_coefficient(a, b)


# 相关系数
def correlation_coefficient(a, b):
    '''
    :param a: 向量1
    :param b: 向量2
    :return: 相关系数
    '''
    a = np.array(a).squeeze()
    b = np.array(b).squeeze()
    if len(a.shape) != 1 or len(b.shape) != 1:
        warnings.warn('数据维度不为1，不执行操作！')
        return None
    return ((a - a.mean()) * (b - b.mean())).mean() / (a.std() * b.std())


# 相关距离
def correlation_distance(a, b):
    '''
    :param a: 向量1
    :param b: 向量2
    :return: 相关距离
    '''
    return 1 - correlation_coefficient(a, b)


# 马氏距离
def mahalanobis_distance(x):
    '''
    :param x: 样本矩阵，形状：[样本个数,特征维数]
    :return: 不同样本间的马氏距离
    '''
    x = np.array(x)
    if len(x.shape) != 2:
        warnings.warn('数据维度不为2，不执行操作！')
        return None
    m, n = x.shape
    S_inv = np.linalg.pinv(np.cov(x))
    D = np.zeros([m, m])
    for i in range(m):
        for j in range(i + 1, m):
            D[i, j] = np.sqrt((x[i] - x[j]).reshape(1, n).dot(S_inv).dot((x[i] - x[j]).reshape(n, 1)))
            D[j, i] = D[i, j]
    return D


# 计算不同样本之间的距离，输入：[m,n]，输出[m,m]，其中dist[i,j]为x[i]到x[j]的距离
def distance_matrix(x):
    '''
    :param x: 样本矩阵，形状：[样本个数,特征维数]
    :return: 不同样本之间的距离
    '''
    m = x.shape[0]
    distance = np.zeros([m, m])
    for i in range(m):
        for j in range(m - 1, i, -1):
            distance[i][j] = np.linalg.norm(x[i] - x[j], 2)
            distance[j][i] = distance[i][j]
    return distance



# Copyright 2018 Zhao Xingyu & An Yuexuan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
import numpy as np
import warnings
import random
from sklearn.manifold import TSNE, MDS
from sklearn.decomposition import PCA, KernelPCA
import matplotlib.pyplot as plt
from scipy import stats
import pandas as pd
from sklearn.preprocessing import minmax_scale
import keras


# 生成样本
def generate_sample(x, y, N_classes=None, sample_num=None, have_correct=True):
    '''
    :param x: 特征
    :param y: 标签
    :param N_classes: 类别个数
    :param sample_num:  样本个数
    :param have_correct: 是否必包括正确标记
    :return: 生成的特征，生成的标签，是否为该类
    '''
    x, y = np.array(x), np.array(y).astype(np.int32)
    if N_classes is None:
        N_classes = y.max() + 1
    if len(y.shape) == 1:
        y = keras.utils.to_categorical(y, N_classes)
    if sample_num is None:
        sample_num = N_classes
    assert sample_num > 0  # 生成的样本必须大于0个
    assert len(y.shape) == 2  # 转化后的标记的维度必须为2
    data, label, is_True = [], [], []
    for i in range(x.shape[0]):
        samples = list(np.random.choice(N_classes, size=sample_num, replace=True))
        if have_correct:
            samples.append(np.argmax(y[i]))
            samples = samples[1:]
        for j in range(sample_num):
            data.append(x[i])
            label.append(samples[j])
            if y[i, samples[j]] == 1:
                is_True.append(True)
            else:
                is_True.append(False)
    return np.array(data), np.array(label), np.array(is_True)


# 生成迭代器
def data_iter(data, label, batch_size, is_True=None):
    '''
    :param data: 样本数据
    :param label: 样本标签
    :param batch_size: 批大小
    :param is_True: 标记是否为真
    :return: 迭代器
    '''
    data, label = np.array(data), np.array(label)
    n_samples = data.shape[0]
    idx = list(range(n_samples))
    random.shuffle(idx)
    if is_True is not None:
        is_True = np.array(is_True)
        for i in range(0, n_samples, batch_size):
            j = np.array(idx[i:min(i + batch_size, n_samples)])
            yield np.take(data, j, 0), np.take(label, j, 0), np.take(is_True, j, 0)
    else:
        for i in range(0, n_samples, batch_size):
            j = np.array(idx[i:min(i + batch_size, n_samples)])
            yield np.take(data, j, 0), np.take(label, j, 0)


# label转one-hot
def to_categorical(label, num_classes=None):
    '''
    :param label: 样本标签
    :param num_classes: 总共的类别数
    :return: 标签的one-hot编码
    '''
    label = np.array(label, dtype='int')
    if num_classes is not None:
        assert num_classes > label.max()  # 类别数量错误
    else:
        num_classes = label.max() + 1
    if len(label.shape) == 1:
        y = np.eye(num_classes, dtype='int64')[label]
        return y
    elif len(label.shape) == 2 and label.shape[1] == 1:
        y = np.eye(num_classes, dtype='int64')[label.squeeze()]
        return y
    else:
        warnings.warn('Warning: one_hot_to_label do not work')
        return label


# one-hot转label
def one_hot_to_label(y):
    '''
    :param y: one-hot编码
    :return: 标签
    '''
    y = np.array(y)
    if len(y.shape) == 2 and y.max() == 1 and y.sum(1).mean() == 1:
        return y.argmax(1)
    else:
        warnings.warn('Warning: one_hot_to_label do not work')
        return y


# 画出样本数据示意图
def plot(x, y, method='t-SNE'):
    '''
    :param x: 数据
    :param y: 标签
    :param method: 可视化方法，包括['t-SNE','PCA','MDS']
    :return: None
    '''
    x, y = check_data_target(x, y)
    if method == 't-SNE':
        x = TSNE(n_components=2).fit_transform(x)
    elif method == 'PCA' or method == 'pca':
        x = PCA(n_components=2).fit_transform(x)
    elif method == 'MDS' or method == 'mds':
        x = MDS(n_components=2).fit_transform(x)
    else:
        warnings.warn('Wrong method!')
        return
    for i in range(y.max() + 1):
        plt.scatter(x[y == i][:, 0], x[y == i][:, 1], label='class %d' % i)
    plt.legend()
    plt.show()


# 画出分类边界图
def plot_classifier(classifier, x, y, method='tsne', figsize=None, save_path=None):
    '''
    :param classifier: 分类器
    :param x: 数据
    :param y: 标签
    :param method: 数据预处理方法，['tsne','mds','pca','kernelpca']
    :param figsize: 图片大小，如果为None，则为默认大小
    :param save_path: 图片保存路径，如果为None，则不保存
    :return: None
    '''
    if x.shape[1] > 2:
        if method == 'tsne':
            x = TSNE(2).fit_transform(x)
        elif method == 'mds':
            x = MDS(2).fit_transform(x)
        elif method == 'kernelpca':
            x = KernelPCA(2).fit_transform(x)
        else:
            x = PCA(2).fit_transform(x)
    x1_min, x1_max = np.min(x[:, 0]) - 1, np.max(x[:, 0]) + 1
    x2_min, x2_max = np.min(x[:, 1]) - 1, np.max(x[:, 1]) + 1
    step_size = 0.01
    x1_values, x2_values = np.meshgrid(np.arange(x1_min, x1_max, step_size), np.arange(x2_min, x2_max, step_size))
    classifier.fit(x, y)
    mesh_output = classifier.predict(np.c_[x1_values.ravel(), x2_values.ravel()])
    mesh_output = mesh_output.reshape(x1_values.shape)
    if figsize is not None:
        plt.figure(figsize=figsize)
    else:
        plt.figure()
    plt.plot(x[y == 0][:, 0], x[y == 0][:, 1], 'bs')
    plt.plot(x[y == 1][:, 0], x[y == 1][:, 1], 'g^')
    plt.pcolormesh(x1_values, x2_values, mesh_output, cmap=plt.cm.Paired)
    plt.scatter(x[:, 0], x[:, 1], c=y, s=80, linewidths=1, cmap=plt.cm.Paired)
    plt.xlim(x1_values.min(), x1_values.max())
    plt.ylim(x2_values.min(), x2_values.max())
    plt.xticks((np.arange(int(np.min(x[:, 0]) - 1), int(np.max(x[:, 0]) + 1), 1)))
    plt.yticks((np.arange(int(np.min(x[:, 1]) - 1), int(np.max(x[:, 1]) + 1), 1)))
    if save_path is not None:
        plt.savefig(fname=save_path, bbox_inches='tight', dpi=300)
    plt.show()


# def plot_classifier(classifier, x, y, axes=[-1.5, 2.5, -1, 1.5]):
#     if x.shape[1]>2:
#         x = PCA(2).fit_transform(x)
#     classifier.fit(x, y)
#     plt.plot(x[y == 0][:, 0], x[y == 0][:, 1], 'bs')
#     plt.plot(x[y == 1][:, 0], x[y == 1][:, 1], 'g^')
#     plt.grid()
#     plt.axis(axes)
#     x0s = np.linspace(axes[0], axes[1], 200)
#     x1s = np.linspace(axes[2], axes[3], 200)
#     x0, x1 = np.meshgrid(x0s, x1s)
#     x = np.c_[x0.ravel(), x1.ravel()]
#     y_pred = classifier.predict(x)
#     y_pred = y_pred.reshape(x0.shape)
#     plt.contourf(x0, x1, y_pred, cmap=plt.cm.brg, alpha=0.2)
#     plt.show()


# 检查数据和标签的形状
def check_data_target(x, y):
    '''
    :param x: 数据
    :param y: 标签
    :return: 数据、标签
    '''
    x, y = np.array(x), np.array(y)
    assert x.ndim == 2  # 数据形状错误
    assert y.ndim == 1 or (y.ndim == 2 and y.shape[1] == 1)  # 标签形状错误
    if y.ndim == 2 and y.shape[1] == 1:
        y = y.squeeze()
    return x, y


# Friedman检验
def Friedman_test(x, alpha=0.05, ranked=False, use_f_distribution=False, verbose=False):
    '''
    :param x: 各个算法在不同数据集上的得分或排序，形状为[数据集个数,算法个数]
    :param alpha: 显著性水平
    :param ranked: 输入的数据是否为排序
    :param use_f_distribution: 是否使用改进的Friedman检测
    :param verbose: 当输入数据为得分时，是否打印排序结果
    :return: 各算法的排序
    '''
    x = np.array(x) + 0.
    n_datasets, n_algorithms = x.shape[0], x.shape[1]
    if not ranked:  # 输入为得分
        for i in range(n_datasets):  # 对于第i个数据集
            rank_list = np.zeros([n_algorithms])  # 不同算法的排名
            score = x[i].copy()
            chuli = 0
            while chuli != n_algorithms:
                M = np.max(score)
                score_equal = []
                for j in range(n_algorithms):
                    if score[j] == M:
                        score_equal.append(j)
                rank_list[score_equal] = np.sum(np.arange(chuli + 1, chuli + 1 + len(score_equal))) / len(score_equal)
                score[score_equal] = -np.inf
                x[i] = rank_list.copy()
                chuli += len(score_equal)
        if verbose:
            print('输入得分排名为：')
            print(x)
    R = np.mean(x, axis=0)
    Tao = 12 * n_datasets / n_algorithms / (n_algorithms + 1) * np.sum(np.square(R - (n_algorithms + 1) / 2))
    if use_f_distribution:  # 使用改进的Friedman检测
        F = stats.f.isf(q=alpha, dfn=(n_algorithms - 1), dfd=int(n_algorithms - 1) * (n_datasets - 1))
        Tao = (n_datasets - 1) * Tao / (n_datasets * (n_algorithms - 1) - Tao)
        if Tao > F:
            print('Tao值为%.4f，显著性水平为%.4f的F分布值为%.4f，有显著区别' % (Tao, alpha, F))
        else:
            print('Tao值为%.4f，显著性水平为%.4f的F分布值为%.4f，无显著区别' % (Tao, alpha, F))
    else:  # 使用传统的Friedman检测
        Chi2 = stats.chi2.isf(q=alpha, df=n_algorithms - 1)
        if Tao > Chi2:
            print('Tao值为%.4f，显著性水平为%.4f的卡方分布值为%.4f，有显著区别' % (Tao, alpha, Chi2))
        else:
            print('Tao值为%.4f，显著性水平为%.4f的卡方分布值为%.4f，无显著区别' % (Tao, alpha, Chi2))
    return x


# t检验
def t_test(x1=None, x2=None, alpha=0.05, from_stats=False, mean1=None, std1=None, nobs1=None, mean2=None, std2=None,
           nobs2=None):
    '''
    :param x1: 第一组准确率,list或numpy.array
    :param x2: 第二组准确率,list或numpy.array
    :param alpha:显著性水平
    :param from_stats: 输入的参数是否为统计学参数
    :param mean1: 第一组样本的均值
    :param std1: 第一组样本的方差
    :param nobs1: 第一组样本个数
    :param mean2: 第二组样本的均值
    :param std2: 第二组样本的方差
    :param nobs2: 第二组样本个数
    :return: 显著程度
    '''
    if from_stats:
        std1, std2 = np.sqrt(nobs1 / (nobs1 - 1)) * std1, np.sqrt(nobs2 / (nobs2 - 1)) * std2
        statistic, pvalue = stats.ttest_ind_from_stats(mean1=mean1, std1=std1, nobs1=nobs1, mean2=mean2, std2=std2,
                                                       nobs2=nobs2)
    else:
        x1, x2 = np.array(x1), np.array(x2)
        statistic, pvalue = stats.levene(x1, x2)
        print(pvalue)
        if pvalue > 0.05:
            equal_val = True
        else:
            equal_val = False
        statistic, pvalue = stats.ttest_ind(x1, x2, equal_var=equal_val)
    if pvalue > alpha:
        print('pvalue值为%.4f，显著性水平为%.4f，无显著区别' % (pvalue, alpha))
    else:
        print('pvalue值为%.4f，显著性水平为%.4f，有显著区别' % (pvalue, alpha))
    return pvalue


# 计算Gram矩阵
def Gram(x):
    '''
    :param x: 输入向量
    :return: 输入向量对应的Gram矩阵
    '''
    x = np.array(x).squeeze()
    assert len(x.shape) == 1  # 样本维度不符
    return x.reshape(-1, 1) * x


# pandas DataFrame数据预处理
def data_preprocessing(data):
    '''
    :param data: Pandas DataFrame
    :return: 处理后的numpy二维数组
    '''
    data = pd.get_dummies(data)
    data = data.fillna(data.mean())
    return minmax_scale(np.array(data))



# -*- coding: utf-8 -*-
# Copyright 2018 Zhao Xingyu & An Yuexuan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-

import numpy as np


# 粒子群算法
class PSO:
    def __init__(self, func=None, param_len=1, size=50, c1=2., c2=2., x_min=-10., x_max=10., v_min=-0.5, v_max=0.5,
                 r1=None, r2=None):
        self.func = func  # 计算适应性系数的方法
        self.param_len = param_len  # 参数个数
        self.size = size  # 有多少微粒
        self.c1 = c1  # 认知系数
        self.c2 = c2  # 社会系数
        self.history = []  # 每次迭代的最优值
        # 最小位移
        if type(x_min) != list:
            self.x_min = [x_min] * self.param_len
        else:
            assert len(x_min) == self.param_len  # 参数个数错误
            self.x_min = x_min
        # 最大位移
        if type(x_max) != list:
            self.x_max = [x_max] * self.param_len
        else:
            assert len(x_max) == self.param_len  # 参数个数错误
            self.x_max = x_max
        # 最小速度
        if type(v_min) != list:
            self.v_min = [v_min] * self.param_len
        else:
            assert len(v_min) == self.param_len  # 参数个数错误
            self.v_min = v_min
        # 最大速度
        if type(v_max) != list:
            self.v_max = [v_max] * self.param_len
        else:
            assert len(v_max) == self.param_len  # 参数个数错误
            self.v_max = v_max
        self.r1 = r1  # 随机数1
        self.r2 = r2  # 随机数2
        self.x = None  # 位移
        self.v = None  # 速度
        self.best_all_x = None  # 全局最优位置
        self.best_all_score = None  # 全局最优分数
        self.best_each_x = None  # 局部最优位置
        self.best_each_score = None  # 局部最优分数
        self._init_fit()

    def _init_fit(self):
        self.x = np.zeros([self.size, self.param_len])  # 形状为[微粒个数,参数个数]
        self.v = np.zeros([self.size, self.param_len])
        for i in range(self.size):
            for j in range(self.param_len):
                self.x[i][j] = np.random.uniform(self.x_min[j], self.x_max[j])
                self.v[i][j] = np.random.uniform(self.v_min[j], self.v_max[j])
        self.best_all_x = np.zeros(self.param_len)  # 全局最优位置
        self.best_all_score = -np.inf  # 全局最优分数
        self.best_each_x = self.x.copy()  # 局部最优位置
        self.best_each_score = np.full(self.size, -np.inf)  # 局部最优分数

    def solve(self, epoch=50, verbose=False):
        self.history = []
        r1 = self.r1
        r2 = self.r2
        for _ in range(epoch):  # 一共迭代_次
            # 配置随机变量
            if r1 is None:
                r1 = np.random.uniform(0, 1)
            if r2 is None:
                r2 = np.random.uniform(0, 1)
            # 计算适应度
            for i in range(self.size):  # 对于第i个微粒
                fitness = self.func(*self.x[i])
                # 更新局部最优值、局部最优位置
                if fitness > self.best_each_score[i]:
                    self.best_each_score[i] = fitness
                    self.best_each_x[i] = self.x[i].copy()
                if fitness > self.best_all_score:
                    self.best_all_score = fitness
                    self.best_all_x = self.x[i].copy()
            # 更新微粒的速度和位置
            self.v = self.v + self.c1 * r1 * (self.best_each_x - self.x) + self.c2 * r2 * (self.best_all_x - self.x)
            for j in range(self.param_len):
                self.v[:, j] = np.clip(self.v[:, j], self.v_min[j], self.v_max[j])
            self.x = self.x + self.v
            for j in range(self.param_len):
                self.x[:, j] = np.clip(self.x[:, j], self.x_min[j], self.x_max[j])
            if verbose:
                # print('已完成第%i次寻找,最优参数值为' % (_ + 1), self.best_all_x, '目前最优适合度为%.4f' % self.best_all_score)
                print('Number of iterations: %i.' % (_ + 1), 'Best fitness: %.4f' % self.best_all_score)
            self.history.append(self.best_all_score)
        self.history = np.array(self.history)
        return self.best_all_x



# -*- coding: utf-8 -*-
# Copyright 2018 Zhao Xingyu & An Yuexuan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
import numpy as np


# 人工鱼群算法
class AFSA:
    def __init__(self, func=None, param_len=1, size=50, x_min=-10., x_max=10., visual=1., step=0.5, delta=1,
                 try_number=5):
        self.func = func  # 计算适应性系数的方法
        self.param_len = param_len  # 参数个数
        self.size = size  # 有多少鱼
        self.history = []  # 每次迭代的最优值
        # 最小位移
        if type(x_min) != list:
            self.x_min = [x_min] * self.param_len
        else:
            assert len(x_min) == self.param_len  # 参数个数错误
            self.x_min = x_min
        # 最大位移
        if type(x_max) != list:
            self.x_max = [x_max] * self.param_len
        else:
            assert len(x_max) == self.param_len  # 参数个数错误
            self.x_max = x_max
        self.visual = visual  # 鱼的视野
        self.step = step  # 鱼的步长
        self.delta = delta  # 拥挤因子
        self.try_number = try_number  # 觅食行为的尝试次数
        self.x = None  # 位移
        self.score = None  # 每条鱼的分数
        self.best_all_x = None  # 全局最优位置
        self.best_all_score = None  # 全局最优分数
        self._init_fit()

    def _init_fit(self):
        self.x = np.zeros([self.size, self.param_len])  # 形状为[鱼的个数,参数个数]
        self.score = np.zeros(self.size)
        for i in range(self.size):
            for j in range(self.param_len):
                self.x[i][j] = np.random.uniform(self.x_min[j], self.x_max[j])
            self.score[i] = self.func(*self.x[i])
        self.best_all_score = np.max(self.score)  # 全局最优分数
        self.best_all_x = self.x[np.argmax(self.score)]  # 全局最优位置

    def solve(self, epoch=50, verbose=False):
        self.history = []
        for _ in range(epoch):  # 一共迭代_次
            for i in range(self.size):  # 对于第i条鱼
                # 检测周围有没有鱼
                fishes = []
                fish_idx = []
                fish_num = 0
                for j in range(self.size):  # 搜索附近有多少鱼
                    if np.sum(np.square(self.x[i] - self.x[j])) < self.visual:
                        fish_num += 1
                        fishes.append(self.x[j])
                        fish_idx.append(j)
                if fish_num > 1:  # 如果周围有鱼
                    fishes = np.array(fishes)
                    # 聚群行为
                    centre = np.mean(fishes, 0)
                    if self.func(*centre) / fish_num > self.delta * self.score[i]:
                        self.x[i] += (centre - self.x[i]) / (np.sum(
                            np.square(centre - self.x[i])) + 10e-8) * self.step * np.random.rand()
                        self.x[i] = np.clip(self.x[i], self.x_min, self.x_max)
                        self.score[i] = self.func(*self.x[i])
                        # 更新局部最优值、局部最优位置
                        if self.score[i] > self.best_all_score:
                            self.best_all_score = self.score[i]
                            self.best_all_x = self.x[i].copy()
                        continue
                    # 追尾行为
                    best_index = fish_idx[np.argmax(self.score[fish_idx])]
                    if self.score[best_index] / fish_num > self.delta * self.score[i]:
                        self.x[i] += (self.x[best_index] - self.x[i]) / np.sum(
                            np.square(self.x[best_index] - self.x[i])) * self.step * np.random.rand()
                        self.x[i] = np.clip(self.x[i], self.x_min, self.x_max)
                        self.score[i] = self.func(*self.x[i])
                        # 更新局部最优值、局部最优位置
                        if self.score[i] > self.best_all_score:
                            self.best_all_score = self.score[i]
                            self.best_all_x = self.x[i].copy()
                        continue
                # 觅食行为
                find = False
                x = None
                for j in range(self.try_number):
                    x = self.x[i] + self.visual * np.random.uniform(-1, 1, self.param_len)
                    x = np.clip(x, self.x_min, self.x_max)
                    if self.func(*x) > self.score[i]:
                        find = True
                    break
                if find is True:
                    self.x[i] += (x - self.x[i]) / np.sum(np.square(x - self.x[i])) * self.step * np.random.rand()
                else:
                    self.x[i] += self.visual * np.random.uniform(-1, 1)
                self.x[i] = np.clip(self.x[i], self.x_min, self.x_max)
                self.score[i] = self.func(*self.x[i])
                # 更新局部最优值、局部最优位置
                if self.score[i] > self.best_all_score:
                    self.best_all_score = self.score[i]
                    self.best_all_x = self.x[i].copy()
            if verbose:
                # print('已完成第%i次寻找,最优参数值为' % (_ + 1), self.best_all_x, '目前最优适合度为%.4f' % self.best_all_score)
                print('Number of iterations: %i. The optimal parameters:' % (_ + 1), self.best_all_x,
                      'Best fitness: %.4f' % self.best_all_score)
            self.history.append(self.best_all_score)
        self.history = np.array(self.history)
        return self.best_all_x



# Use Q-Learning for solving the FrozenLake problem
from ailearn.RL.TabularRL import QLearning
from ailearn.RL.Environment import FrozenLake

env = FrozenLake(4)
agent = QLearning(env.n_actions, env.n_states)
MAX_EPISODES = 200

for episode in range(MAX_EPISODES):
    s = env.reset()
    total_reward = 0
    while True:
        a = agent.choose_action(s)
        s_, r, done, _ = env.step(a)
        agent.learn(s, a, r, s_, done)
        s = s_
        total_reward = total_reward + r
        if done:
            break
    print('episode:', episode, 'total_reward:', total_reward)



# -*- coding: utf-8 -*-
# Copyright 2018 Zhao Xingyu & An Yuexuan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
import numpy as np


# relu激活函数
def relu(x):
    x = np.array(x)
    return np.maximum(0, x)


# tanh激活函数
def tanh(x):
    x = np.array(x)
    return np.tanh(x)


# sigmoid激活函数
def sigmoid(x):
    x = np.array(x)
    return 1 / (1 + np.exp(-x))


# softmax激活函数
def softmax(x):
    x = np.array(x)
    assert len(x.shape) == 1 or len(x.shape) == 2
    if len(x.shape) == 1:
        x = x - x.max()
        x = np.exp(x)
        return x / x.sum()
    else:
        x = x - x.max(1, keepdims=True)
        x = np.exp(x)
        return x / x.sum(1, keepdims=True)


# linear激活函数
def linear(x):
    x = np.array(x)
    return x


# 阈值激活函数
def threshold(x, threshold=0):
    x = np.array(x)
    out = np.zeros_like(x, dtype=np.float)
    out[x >= threshold] = 1
    return out


# arctan激活函数
def arctan(x):
    x = np.array(x)
    return np.arctan(x)


# leaky relu
def leaky_relu(x, alpha=0.1):
    x = np.array(x, dtype=np.float)
    x[x < 0] = (x * alpha)[x < 0]
    return x


# prelu激活函数
def prelu(x, p):
    x = np.array(x, dtype=np.float)
    x[x < 0] = (x * p)[x < 0]
    return x


# elu激活函数
def elu(x, alpha=0.1):
    x = np.array(x, dtype=np.float)
    x[x < 0] = (alpha * (np.exp(x) - 1))[x < 0]
    return x


# softplus激活函数
def softplus(x):
    x = np.array(x)
    return np.log(1 + np.exp(x))


# bent identity
def bent_identity(x):
    x = np.array(x)
    return (np.sqrt(np.square(x) + 1) - 1) * 0.5 + x


# Soft Exponential
def soft_exponential(x, p):
    x = np.array(x, dtype=np.float)
    x[p < 0] = (-np.log(np.maximum(1 - p[p < 0] * (x[p < 0] + p[p < 0]), 1e-7)) / p[p < 0])
    x[p == 0] = 0
    x[p > 0] = ((np.exp(p * x) - 1) / p + p)[p > 0]
    return x


# Sinusoid
def sin(x):
    x = np.array(x)
    return np.sin(x)


# Sinc
def sinc(x):
    x = np.array(x, dtype=np.float)
    out = np.ones_like(x, dtype=np.float)
    out[x != 0] = np.sin(x[x != 0]) / x[x != 0]
    return out


# Gaussian
def guassian(x):
    x = np.array(x)
    return np.exp(-np.square(x))



# Copyright 2018 Zhao Xingyu & An Yuexuan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
from keras.preprocessing.text import Tokenizer
from keras.preprocessing.sequence import pad_sequences


# 本文列表转化成index列表
def texts_to_sequences(texts, num_words=None, filters='!"#$%&()*+,-./:;<=>?@[\\]^_`{|}~\t\n', lower=True, split=' ',
                       char_level=False, oov_token=None, document_count=0, is_padding=True, maxlen=None, dtype='int32',
                       padding='pre', truncating='pre', value=0):
    '''
    :param texts: 待转换的文本
    :param num_words: None或整数，处理的最大单词数量。若被设置为整数，则分词器将被限制为待处理数据集中最常见的num_words个单词
    :param filters: 需要滤除的字符的列表或连接形成的字符串
    :param lower: 布尔值，是否将序列设为小写形式
    :param split: 字符串，单词的分隔符，如空格
    :param char_level: 如果为 True, 每个字符将被视为一个标记
    :param oov_token: out-of-vocabulary，如果给定一个string作为这个oov token的话，就将这个string也加到word_index，也就是从word到index 的映射中，用来代替那些字典上没有的字。
    :param document_count: 整数。分词器被训练的文档（文本或者序列）数量。仅在调用fit_on_texts或fit_on_sequences之后设置。
    :param is_padding: 布尔值，是否对序列进行填充
    :param maxlen: 序列的最大长度
    :param dtype: numpy数组的数据类型
    :param padding: 'pre'或'post'，在前面或后面填充
    :param truncating: 'pre'或'post'，在起始或结尾截断
    :param value: 填充值
    :return: 转化后的index序列，index到word的字典，word到index的字典
    '''
    tokenizer = Tokenizer(num_words=num_words, filters=filters, lower=lower, split=split, char_level=char_level,
                          oov_token=oov_token, document_count=document_count)
    tokenizer.fit_on_texts(texts)
    sequences = tokenizer.texts_to_sequences(texts)
    index2word = tokenizer.index_word
    word2index = tokenizer.word_index
    if is_padding:
        sequences = pad_sequences(sequences, maxlen=maxlen, dtype=dtype, padding=padding, truncating=truncating,
                                  value=value)
    return sequences, index2word, word2index



# Copyright 2018 Zhao Xingyu & An Yuexuan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
# 智能算法的若干不同的评价函数
import numpy as np


# Ackley函数
class Ackley:
    def __init__(self, a=20, b=0.2, c=2 * np.pi):
        self.min = -32.768
        self.max = 32.768
        self.a = a
        self.b = b
        self.c = c

    def func(self, *x):
        x = np.array(x)
        mean1 = np.mean(np.square(x))
        mean2 = np.mean(np.cos(self.c * x))
        result = -self.a * np.exp(-self.b * np.sqrt(mean1)) - np.exp(mean2) + self.a + np.e
        return -result


# Griewank函数
class Griewank:
    def __init__(self):
        self.min = -600
        self.max = 600

    def func(self, *x):
        x = np.array(x)
        first = np.sum(np.square(x)) / 4000
        second = 1
        for i in range(x.shape[0]):
            second *= np.cos(x[i] / np.sqrt(i + 1))
        result = first - second + 1
        return -result


# Rastrigin函数
class Rastrigin:
    def __init__(self):
        self.min = -5.12
        self.max = 5.12

    def func(self, *x):
        x = np.array(x)
        sum = np.sum(np.square(x) - 10 * np.cos(2 * np.pi * x))
        result = 10 * x.shape[0] + sum
        return -result


# Rosenbrock函数
class Rosenbrock:
    def __init__(self):
        self.min = -5
        self.max = 10

    def func(self, x1, x2):
        result = np.square(1 - x1) + 100 * np.square(x2 - np.square(x1))
        return -result


# Schwefel函数
class Schwefel:
    def __init__(self):
        self.min = -500
        self.max = 500

    def func(self, *x):
        x = np.array(x)
        sum = np.sum(x * np.sin(np.sqrt(np.abs(x))))
        result = 418.9829 * x.shape[0] - sum
        return -result


# Sphere函数
class Sphere:
    def __init__(self):
        self.min = -100
        self.max = 100

    def func(self, *x):
        x = np.array(x)
        result = np.sum(np.square(x))
        return -result


# Schwefel's Problem 1.2
class Schwefel_1:
    def __init__(self):
        self.min = -100
        self.max = 100

    def func(self, *x):
        x = np.array(x)
        a = np.zeros_like(x)
        for i in range(x.shape[0]):
            a[i] = x[:i].sum() ** 2
        return -a.sum()


# Schwefel's Problem 2.22
class Schwefel_2:
    def __init__(self):
        self.min = -10
        self.max = 10

    def func(self, *x):
        x = np.array(x)
        return -np.sum(np.abs(x)) - np.prod(np.abs(x))


# Branins函数
class Branins:
    def __init__(self):
        self.min = [-5, 0]
        self.max = [10, 15]

    def func(self, x1, x2):
        result = (x2 - (5.1 / (4 * np.pi ** 2)) + 5 / np.pi * x1 - 6) ** 2 + 10 * (1 - 1 / 8 / np.pi) * np.cos(x1) + 10
        return -result



# Copyright 2018 Zhao Xingyu & An Yuexuan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

with open('README.rst', 'r', encoding='utf8') as f:
    long_description = f.read()
setup(
    name='ailearn',
    version='0.2.1.7',
    description='A lightweight package for artificial intelligence',
    long_description=long_description,
    author='ZHAO Xingyu; AN Yuexuan',
    author_email='757008724@qq.com',
    license='Apache License, Version 2.0',
    url='http://github.com/axi345/ailearn',
    packages=find_packages(),
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Information Technology',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Environment :: Console',
    ],
    zip_safe=False,
    install_requires=['numpy', 'pandas', 'sklearn', 'matplotlib', 'scipy', 'keras', 'bs4', 'requests'],
)



import numpy as np
import torch
import random


# 设置随机数种子
def set_seed(seed):
    torch.manual_seed(seed)  # cpu
    torch.cuda.manual_seed(seed)  # gpu
    torch.backends.cudnn.deterministic = True  # cudnn
    np.random.seed(seed)  # numpy
    random.seed(seed)  # random and transforms



# -*- coding: utf-8 -*-
# Copyright 2018 Zhao Xingyu & An Yuexuan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
import numpy as np


# 萤火虫算法
class FA:
    def __init__(self, func=None, param_len=1, size=50, x_min=-10., x_max=10., beta=2, alpha=0.5, gamma=0.9):
        self.func = func  # 计算适应性系数的方法
        self.param_len = param_len  # 参数个数
        self.size = size  # 有多少萤火虫
        self.history = []  # 每次迭代的最优值
        # 最小位移
        if type(x_min) != list:
            self.x_min = [x_min] * self.param_len
        else:
            assert len(x_min) == self.param_len  # 参数个数错误
            self.x_min = x_min
        # 最大位移
        if type(x_max) != list:
            self.x_max = [x_max] * self.param_len
        else:
            assert len(x_max) == self.param_len  # 参数个数错误
            self.x_max = x_max
        self.alpha = alpha  # 步长
        self.beta = beta  # 最大吸引度
        self.gamma = gamma  # 光强吸收系数
        self.x = None  # 位移
        self.score = None  # 每个萤火虫的分数
        self.best_all_x = None  # 全局最优位置
        self.best_all_score = None  # 全局最优分数
        self._init_fit()

    def _init_fit(self):
        self.x = np.zeros([self.size, self.param_len])  # 形状为[萤火虫的个数,参数个数]
        self.score = np.zeros(self.size)
        for i in range(self.size):
            for j in range(self.param_len):
                self.x[i][j] = np.random.uniform(self.x_min[j], self.x_max[j])
            self.score[i] = self.func(*self.x[i])
        self.best_all_score = np.max(self.score)  # 全局最优分数
        self.best_all_x = self.x[np.argmax(self.score)]  # 全局最优位置

    def solve(self, epoch=50, verbose=False):
        self.history = []
        for _ in range(epoch):  # 一共迭代_次
            # 计算距离
            distance = np.zeros([self.size, self.size])
            for i in range(self.size):
                for j in range(self.size - 1, i, -1):
                    distance[i][j] = np.linalg.norm(self.x[i] - self.x[j], 2)
                    distance[j][i] = distance[i][j]
            for i in range(self.size):  # 对于第i个萤火虫
                # 对于最亮的萤火虫，做随机运动
                if np.argmax(self.score) == i:
                    self.x[i] += self.alpha * np.random.uniform(-0.5, 0.5)
                    self.x[i] = np.clip(self.x[i], self.x_min, self.x_max)
                    self.score[i] = self.func(*self.x[i])
                    # 更新局部最优值、局部最优位置
                    if self.score[i] > self.best_all_score:
                        self.best_all_score = self.score[i]
                        self.best_all_x = self.x[i].copy()
                    continue
                lightness = np.zeros(self.size)
                # 查找对自己相对亮度最大的萤火虫
                for j in range(self.size):
                    if j == i:
                        lightness[j] = -np.inf
                    else:
                        lightness[j] = self.score[j] * np.exp(-self.gamma * distance[i][j])
                idx = np.argmax(lightness)
                # 更新萤火虫的位置
                self.x[i] += self.beta * np.exp(-self.gamma * np.square(distance[i][idx])) * (
                        self.x[idx] - self.x[i]) + self.alpha * np.random.uniform(-0.5, 0.5)
                self.score[i] = self.func(*self.x[i])
                self.x[i] = np.clip(self.x[i], self.x_min, self.x_max)
                # 更新局部最优值、局部最优位置
                if self.score[i] > self.best_all_score:
                    self.best_all_score = self.score[i]
                    self.best_all_x = self.x[i].copy()
            if verbose:
                # print('已完成第%i次寻找,最优参数值为' % (_ + 1), self.best_all_x, '目前最优适合度为%.4f' % self.best_all_score)
                print('Number of iterations: %i. The optimal parameters:' % (_ + 1), self.best_all_x,
                      'Best fitness: %.4f' % self.best_all_score)
            self.history.append(self.best_all_score)
        self.history = np.array(self.history)
        return self.best_all_x



# -*- coding: utf-8 -*-
# Copyright 2018 Zhao Xingyu & An Yuexuan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
import numpy as np
from ..utils import to_categorical
from .activations import softmax, sigmoid


# softmax交叉熵
def softmax_cross_entropy(out, label):
    # out:神经元的输出值
    # label:实际类别或one-hot编码
    out, label = np.array(out), np.array(label)
    assert len(out.shape) == 2  # 输出形状错误
    assert len(label.shape) == 1 or len(label.shape) == 2  # 标签形状错误
    if len(label.shape) == 1:  # 转化为one-hot编码
        y = to_categorical(label, num_classes=out.shape[1])
    else:
        if label.shape[1] == 1:
            y = to_categorical(label.squeeze(), num_classes=out.shape[1])
        else:
            assert label.max() == 1 and label.sum(1).mean() == 1  # 标签one-hot编码错误
            y = label
    yhat = softmax(out)
    return -np.mean(y * np.log(yhat))


# 交叉熵
def cross_entropy(out, label):
    # out:神经元的输出值
    # label:实际类别或one-hot编码
    yhat, label = np.array(out), np.array(label)
    assert len(out.shape) == 2  # 输出形状错误
    assert len(label.shape) == 1 or len(label.shape) == 2  # 标签形状错误
    if len(label.shape) == 1:  # 转化为one-hot编码
        y = to_categorical(label, num_classes=out.shape[1])
    else:
        if label.shape[1] == 1:
            y = to_categorical(label.squeeze(), num_classes=out.shape[1])
        else:
            assert label.max() == 1 and label.sum(1).mean() == 1  # 标签one-hot编码错误
            y = label
    return -np.mean(y * np.log(yhat))


# 二分类
def sigmoid_binary_cross_entropy(out, label):
    # out:神经元的输出值
    # label:实际类别或one-hot编码
    out, y = np.array(out), np.array(label)
    assert len(out.shape) == 2 and out.shape[1] == 1  # 输出形状错误
    assert len(y.shape) == 1  # 标签形状错误
    yhat = sigmoid(out)
    return -np.mean(y * np.log(yhat) + (1 - y) * np.log(1 - yhat))


# 二分类
def binary_cross_entropy(out, label):
    # out:神经元的输出值
    # label:实际类别或one-hot编码
    yhat, y = np.array(out), np.array(label)
    assert len(yhat.shape) == 2 and out.shape[1] == 1  # 输出形状错误
    assert len(y.shape) == 1  # 标签形状错误
    return -np.mean(y * np.log(yhat) + (1 - y) * np.log(1 - yhat))


# 最小二乘损失
def square_loss(prediction, y):
    # prediction:预测值
    # y:实际值
    prediction, y = np.array(prediction), np.array(y)
    assert (len(prediction.shape) == 2 and prediction.shape[1] == 1) or len(prediction.shape) == 1  # 输出形状错误
    assert len(y.shape) == 1 or (len(y.shape) == 2 and y.shape[1] == 1)  # 真实值形状错误
    return np.sum(np.sum(np.square(prediction.reshape(-1, 1) - y.reshape(-1, 1)), 1))


# 均方误差
def mse(prediction, y):
    # prediction:预测值
    # y:实际值
    prediction, y = np.array(prediction), np.array(y)
    assert (len(prediction.shape) == 2 and prediction.shape[1] == 1) or len(prediction.shape) == 1  # 输出形状错误
    assert len(y.shape) == 1 or (len(y.shape) == 2 and y.shape[1] == 1)  # 真实值形状错误
    return np.mean(np.sum(np.square(prediction.reshape(-1, 1) - y.reshape(-1, 1)), 1))



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import warnings

np.random.seed(1)


class TabularRL(object):
    def __init__(self, n_actions, n_states, learning_rate=0.1, reward_decay=0.9, method='e_greedy', e_greedy=0.01):
        self.n_actions = n_actions
        self.n_states = n_states
        self.lr = learning_rate
        self.gamma = reward_decay
        self.method = method
        self.epsilon = e_greedy
        self.q_table = np.zeros([n_states, n_actions])
        if self.method != 'e_greedy' and self.method != 'Boltzmann':
            warnings.warn('No such method, using e_greedy method.')
            self.method = 'e_greedy'

    def choose_action(self, s):
        if self.method == 'Boltzmann':
            q_value = self.q_table[s, :]
            exp_q_value = np.exp(q_value)
            action_prob = exp_q_value / np.sum(exp_q_value)
            return np.random.choice(self.n_actions, p=action_prob)
        else:
            if np.random.uniform() < self.epsilon:
                a = np.random.choice(self.n_actions)
            else:
                action_list = []
                max = np.max(self.q_table[s, :])
                for i in range(self.n_actions):
                    if self.q_table[s, i] == max:
                        action_list.append(i)
                a = np.random.choice(action_list)
            return a

    def learn(self, *args):
        pass


class QLearning(TabularRL):
    def learn(self, s, a, r, s_, done):
        q_predict = self.q_table[s, a]
        if done:
            q_target = r
        else:
            q_target = r + self.gamma * np.max(self.q_table[s_, :])
        self.q_table[s, a] += self.lr * (q_target - q_predict)


class SarsaLearning(TabularRL):
    def learn(self, s, a, r, s_, a_, done):
        q_predict = self.q_table[s, a]
        if done:
            q_target = r
        else:
            q_target = r + self.gamma * self.q_table[s_, a_]
        self.q_table[s, a] += self.lr * (q_target - q_predict)


class SarsaLambda(TabularRL):
    def __init__(self, n_actions, n_states, learning_rate=0.1, reward_decay=0.9, method='e_greedy', e_greedy=0.01,
                 trace_decay=0.9, style='replacing'):
        super(SarsaLambda, self).__init__(n_actions, n_states, learning_rate, reward_decay, method, e_greedy)
        self.lambda_ = trace_decay
        self.eligibility_trace = self.q_table.copy()
        self.style = style
        if self.style != 'replacing' and self.style != 'accumulating':
            warnings.warn('No such style, using replacing updating method.')
            self.style = 'replacing'

    def learn(self, s, a, r, s_, a_, done):
        q_predict = self.q_table[s, a]
        if done:
            q_target = r
        else:
            q_target = r + self.gamma * self.q_table[s_, a_]
        error = q_target - q_predict
        if self.style == 'replacing':
            self.eligibility_trace[s, :] *= 0
            self.eligibility_trace[s, a] = 1
        else:
            self.eligibility_trace[s, a] += 1
        self.q_table += self.lr * error * self.eligibility_trace
        self.eligibility_trace *= self.gamma * self.lambda_


class QLambda(TabularRL):
    def __init__(self, n_actions, n_states, learning_rate=0.1, reward_decay=0.9, method='e_greedy', e_greedy=0.01,
                 trace_decay=0.9, style='replacing'):
        super(QLambda, self).__init__(n_actions, n_states, learning_rate, reward_decay, method, e_greedy)
        self.lambda_ = trace_decay
        self.eligibility_trace = self.q_table.copy()
        self.style = style
        if self.style != 'replacing' and self.style != 'accumulating':
            warnings.warn('No such style, using replacing updating method.')
            self.style = 'replacing'

    def learn(self, s, a, r, s_, done):
        q_predict = self.q_table[s, a]
        if done:
            q_target = r
        else:
            q_target = r + self.gamma * np.max(self.q_table[s_, :])
        error = q_target - q_predict
        if self.style == 'replacing':
            self.eligibility_trace[s, :] *= 0
            self.eligibility_trace[s, a] = 1
        else:
            self.eligibility_trace[s, a] += 1
        self.q_table += self.lr * error * self.eligibility_trace
        self.eligibility_trace *= self.gamma * self.lambda_


class EnvModel(object):
    def __init__(self, n_actions):
        self.n_actions = n_actions
        self.database = pd.DataFrame(columns=n_actions, dtype=np.object)

    def store_transition(self, s, a, r, s_, done):
        if s not in self.database.index:
            self.database = self.database.append(
                pd.Series(
                    [None] * len(self.n_actions),
                    index=self.database.columns,
                    name=s
                ))
        self.database.set_value(s, a, (r, s_, done))

    def sample_s_a(self):
        s = np.random.choice(self.database.index)
        a = np.random.choice(self.database.ix[s].dropna().index)  # filter out the None value
        return s, a

    def get_r_s_(self, s, a):
        r, s_, done = self.database.ix[s, a]
        return r, s_, done


class DynaQ(TabularRL):
    def __init__(self, n_actions, n_states, learning_rate=0.1, reward_decay=0.9, method='e_greedy', e_greedy=0.01,
                 learning_n=5):
        super(DynaQ, self).__init__(n_actions, n_states, learning_rate, reward_decay, method, e_greedy)
        self.model = EnvModel(list(range(n_actions)))
        self.learning_n = learning_n

    def learn(self, s, a, r, s_, done):
        self.model.store_transition(s, a, r, s_, done)
        q_predict = self.q_table[s, a]
        if done:
            q_target = r
        else:
            q_target = r + self.gamma * np.max(self.q_table[s_, :])
        self.q_table[s, a] += self.lr * (q_target - q_predict)
        for i in range(self.learning_n):
            s, a = self.model.sample_s_a()
            r, s_, done = self.model.get_r_s_(s, a)
            q_predict = self.q_table[s, a]
            if done:
                q_target = r
            else:
                q_target = r + self.gamma * np.max(self.q_table[s_, :])
            self.q_table[s, a] += self.lr * (q_target - q_predict)
